import datetime
from typing import Any, Mapping, Optional, Sequence

import pendulum
from pydantic import BaseModel

from dagster import _check as check
from dagster._annotations import experimental
from dagster._serdes.serdes import (
    FieldSerializer,
    JsonSerializableValue,
    PackableValue,
    SerializableNonScalarKeyMapping,
    UnpackContext,
    WhitelistMap,
    deserialize_value,
    pack_value,
    serialize_value,
    unpack_value,
    whitelist_for_serdes,
)
from dagster._utils.merger import merge_dicts
from dagster._utils.schedules import get_latest_completed_cron_tick

from ..asset_check_spec import AssetCheckKey, AssetCheckSpec
from ..asset_checks import AssetChecksDefinition
from ..asset_selection import AssetSelection
from ..decorators import sensor
from ..run_request import RunRequest
from ..sensor_definition import DefaultSensorStatus, SensorDefinition, SensorEvaluationContext
from .utils import (
    DEADLINE_CRON_METADATA_KEY,
    DEFAULT_FRESHNESS_TIMEZONE,
    FRESHNESS_PARAMS_METADATA_KEY,
    FRESHNESS_TIMEZONE_METADATA_KEY,
    LOWER_BOUND_DELTA_METADATA_KEY,
    ensure_freshness_checks,
    ensure_no_duplicate_asset_checks,
)

DEFAULT_FRESHNESS_SENSOR_NAME = "freshness_checks_sensor"


class EvaluationTimestampsSerializer(FieldSerializer):
    def pack(
        self,
        mapping: Mapping[str, float],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue:
        return pack_value(SerializableNonScalarKeyMapping(mapping), whitelist_map, descent_path)

    def unpack(
        self,
        unpacked_value: JsonSerializableValue,
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> PackableValue:
        return unpack_value(unpacked_value, dict, whitelist_map, context)


@whitelist_for_serdes(
    field_serializers={"evaluation_timestamps_by_check_key": EvaluationTimestampsSerializer}
)
class FreshnessCheckSensorCursor(BaseModel):
    """The cursor for the freshness check sensor."""

    evaluation_timestamps_by_check_key: Mapping[AssetCheckKey, float]

    def get_last_evaluation_timestamp(self, key: AssetCheckKey) -> Optional[float]:
        return self.evaluation_timestamps_by_check_key.get(key)

    @staticmethod
    def empty():
        return FreshnessCheckSensorCursor(evaluation_timestamps_by_check_key={})

    def with_updated_evaluations(
        self, keys: Sequence[AssetCheckKey], evaluation_timestamp: float
    ) -> "FreshnessCheckSensorCursor":
        return FreshnessCheckSensorCursor(
            evaluation_timestamps_by_check_key=merge_dicts(
                self.evaluation_timestamps_by_check_key,
                {key: evaluation_timestamp for key in keys},
            )
        )


@experimental
def build_sensor_for_freshness_checks(
    *,
    freshness_checks: Sequence[AssetChecksDefinition],
    minimum_interval_seconds: Optional[int] = None,
    name: str = DEFAULT_FRESHNESS_SENSOR_NAME,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
) -> SensorDefinition:
    """Builds a sensor which kicks off evaluation of freshness checks.

    The sensor will introspect from the parameters of the passed-in freshness checks how often to
    run them. IE, if the freshness check is based on a cron schedule, the sensor will request one
    run of the check per tick of the cron. If the freshness check is based on a maximum lag, the
    sensor will request one run of the check per interval specified by the maximum lag.

    Args:
        freshness_checks (Sequence[AssetChecksDefinition]): The freshness checks to evaluate.
        minimum_interval_seconds (Optional[int]): The duration in seconds between evaluations of the sensor.
        name (Optional[str]): The name of the sensor. Defaults to "freshness_check_sensor", but a
            name may need to be provided in case of multiple calls of this function.
        default_status (Optional[DefaultSensorStatus]): The default status of the sensor. Defaults
            to stopped.

    Returns:
        SensorDefinition: The sensor that evaluates the freshness of the assets.
    """
    freshness_checks = check.sequence_param(freshness_checks, "freshness_checks")
    ensure_no_duplicate_asset_checks(freshness_checks)
    ensure_freshness_checks(freshness_checks)
    check.invariant(
        check.int_param(minimum_interval_seconds, "minimum_interval_seconds") > 0
        if minimum_interval_seconds
        else True,
        "Interval must be a positive integer.",
    )
    check.str_param(name, "name")

    @sensor(
        name=name,
        minimum_interval_seconds=minimum_interval_seconds,
        description="Evaluates the freshness of targeted assets.",
        asset_selection=AssetSelection.checks(*freshness_checks),
        default_status=default_status,
    )
    def the_sensor(context: SensorEvaluationContext) -> Optional[RunRequest]:
        cursor = (
            deserialize_value(context.cursor, FreshnessCheckSensorCursor)
            if context.cursor
            else FreshnessCheckSensorCursor.empty()
        )
        current_timestamp = pendulum.now("UTC").timestamp()
        check_keys_to_run = []
        for asset_check in freshness_checks:
            for asset_check_spec in asset_check.check_specs:
                prev_evaluation_timestamp = cursor.get_last_evaluation_timestamp(
                    asset_check_spec.key
                )
                if should_run_again(asset_check_spec, prev_evaluation_timestamp, current_timestamp):
                    check_keys_to_run.append(asset_check_spec.key)
        context.update_cursor(
            serialize_value(cursor.with_updated_evaluations(check_keys_to_run, current_timestamp))
        )
        if check_keys_to_run:
            return RunRequest(asset_check_keys=check_keys_to_run)

    return the_sensor


def should_run_again(
    check_spec: AssetCheckSpec, prev_evaluation_timestamp: Optional[float], current_timestamp: float
) -> bool:
    metadata = check_spec.metadata
    if not metadata:
        # This should never happen, but type hinting prevents us from using check.not_none
        check.assert_never(metadata)

    if not prev_evaluation_timestamp:
        return True

    timezone = get_freshness_cron_timezone(metadata) or DEFAULT_FRESHNESS_TIMEZONE
    current_time = pendulum.from_timestamp(current_timestamp, tz=timezone)
    prev_evaluation_time = pendulum.from_timestamp(prev_evaluation_timestamp, tz=timezone)

    deadline_cron = get_freshness_cron(metadata)
    if deadline_cron:
        latest_completed_cron_tick = check.not_none(
            get_latest_completed_cron_tick(deadline_cron, current_time, timezone)
        )
        return latest_completed_cron_tick > prev_evaluation_time

    lower_bound_delta = check.not_none(get_lower_bound_delta(metadata))
    return current_time > prev_evaluation_time + lower_bound_delta


def get_metadata(check_spec: AssetCheckSpec) -> Mapping[str, Any]:
    if check_spec.metadata:
        return check_spec.metadata
    check.assert_never(check_spec.metadata)


def get_freshness_cron(metadata: Mapping[str, Any]) -> Optional[str]:
    return metadata[FRESHNESS_PARAMS_METADATA_KEY].get(DEADLINE_CRON_METADATA_KEY)


def get_freshness_cron_timezone(metadata: Mapping[str, Any]) -> Optional[str]:
    return metadata[FRESHNESS_PARAMS_METADATA_KEY].get(FRESHNESS_TIMEZONE_METADATA_KEY)


def get_lower_bound_delta(metadata: Mapping[str, Any]) -> Optional[datetime.timedelta]:
    float_delta: float = metadata[FRESHNESS_PARAMS_METADATA_KEY].get(LOWER_BOUND_DELTA_METADATA_KEY)
    return datetime.timedelta(seconds=float_delta) if float_delta else None



from collections import defaultdict
from functools import cached_property
from typing import AbstractSet, DefaultDict, Dict, Iterable, Mapping, Optional, Sequence, Set, Union

from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_spec import (
    SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET,
    AssetExecutionType,
    AssetSpec,
)
from dagster._core.definitions.assets import AssetsDefinition, asset_owner_to_str
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.base_asset_graph import (
    AssetKeyOrCheckKey,
    BaseAssetGraph,
    BaseAssetNode,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import ArbitraryMetadataMapping
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.resolved_asset_deps import ResolvedAssetDependencies
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import DEFAULT_GROUP_NAME
from dagster._core.selector.subset_selector import (
    generate_asset_dep_graph,
)


class AssetNode(BaseAssetNode):
    def __init__(
        self,
        key: AssetKey,
        parent_keys: AbstractSet[AssetKey],
        child_keys: AbstractSet[AssetKey],
        assets_def: AssetsDefinition,
        check_keys: AbstractSet[AssetCheckKey],
    ):
        self.key = key
        self.parent_keys = parent_keys
        self.child_keys = child_keys
        self.assets_def = assets_def
        self._check_keys = check_keys

    @property
    def description(self) -> Optional[str]:
        return self.assets_def.descriptions_by_key.get(self.key)

    @property
    def group_name(self) -> str:
        return self.assets_def.group_names_by_key.get(self.key, DEFAULT_GROUP_NAME)

    @property
    def is_materializable(self) -> bool:
        return self.assets_def.is_materializable

    @property
    def is_observable(self) -> bool:
        return self.assets_def.is_observable

    @property
    def is_external(self) -> bool:
        return self.assets_def.is_external

    @property
    def is_executable(self) -> bool:
        return self.assets_def.is_executable

    @property
    def metadata(self) -> ArbitraryMetadataMapping:
        return self.assets_def.metadata_by_key.get(self.key, {})

    @property
    def tags(self) -> Mapping[str, str]:
        return self.assets_def.tags_by_key.get(self.key, {})

    @property
    def owners(self) -> Sequence[str]:
        return [
            asset_owner_to_str(owner) for owner in self.assets_def.owners_by_key.get(self.key, [])
        ]

    @property
    def is_partitioned(self) -> bool:
        return self.assets_def.partitions_def is not None

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        return self.assets_def.partitions_def

    @property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        return self.assets_def.partition_mappings

    @property
    def freshness_policy(self) -> Optional[FreshnessPolicy]:
        return self.assets_def.freshness_policies_by_key.get(self.key)

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        return self.assets_def.auto_materialize_policies_by_key.get(self.key)

    @property
    def auto_observe_interval_minutes(self) -> Optional[float]:
        return self.assets_def.auto_observe_interval_minutes

    @property
    def backfill_policy(self) -> Optional[BackfillPolicy]:
        return self.assets_def.backfill_policy

    @property
    def code_version(self) -> Optional[str]:
        return self.assets_def.code_versions_by_key.get(self.key)

    @property
    def check_keys(self) -> AbstractSet[AssetCheckKey]:
        return self._check_keys

    @property
    def execution_set_asset_keys(self) -> AbstractSet[AssetKey]:
        return (
            {self.key}
            if len(self.assets_def.keys) <= 1 or self.assets_def.can_subset
            else self.assets_def.keys
        )

    @property
    def execution_set_asset_and_check_keys(self) -> AbstractSet[AssetKeyOrCheckKey]:
        if self.assets_def.can_subset:
            return {self.key}
        else:
            return self.assets_def.asset_and_check_keys

    ##### ASSET GRAPH SPECIFIC INTERFACE

    @property
    def execution_type(self) -> AssetExecutionType:
        return self.assets_def.execution_type

    @property
    def io_manager_key(self) -> str:
        return self.assets_def.get_io_manager_key_for_asset_key(self.key)


class AssetGraph(BaseAssetGraph[AssetNode]):
    _assets_defs_by_check_key: Mapping[AssetCheckKey, AssetsDefinition]

    def __init__(
        self,
        asset_nodes_by_key: Mapping[AssetKey, AssetNode],
        assets_defs_by_check_key: Mapping[AssetCheckKey, AssetsDefinition],
    ):
        self._asset_nodes_by_key = asset_nodes_by_key
        self._assets_defs_by_check_key = assets_defs_by_check_key

    @staticmethod
    def normalize_assets(
        assets: Iterable[Union[AssetsDefinition, SourceAsset]],
    ) -> Sequence[AssetsDefinition]:
        """Normalize a mixed list of AssetsDefinition and SourceAsset to a list of AssetsDefinition.

        Normalization includse:

        - Converting any SourceAsset to an AssetDefinition
        - Resolving all relative asset keys (that sometimes specify dependencies) to absolute asset
          keys
        - Creating unexecutable external asset definitions for any keys referenced by asset checks
          or as dependencies, but for which no definition was provided.
        """
        from dagster._core.definitions.external_asset import (
            create_external_asset_from_source_asset,
            external_asset_from_spec,
        )

        # Convert any source assets to external assets
        assets_defs = [
            create_external_asset_from_source_asset(a) if isinstance(a, SourceAsset) else a
            for a in assets
        ]
        all_keys = {k for asset_def in assets_defs for k in asset_def.keys}

        # Resolve all asset dependencies. An asset dependency is resolved when its key is an
        # AssetKey not subject to any further manipulation.
        resolved_deps = ResolvedAssetDependencies(assets_defs, [])
        assets_defs = [
            ad.with_attributes(
                input_asset_key_replacements={
                    raw_key: resolved_deps.get_resolved_asset_key_for_input(ad, input_name)
                    for input_name, raw_key in ad.keys_by_input_name.items()
                }
            )
            for ad in assets_defs
        ]

        # Create unexecutable external assets definitions for any referenced keys for which no
        # definition was provided.
        all_referenced_asset_keys = {
            key for assets_def in assets_defs for key in assets_def.dependency_keys
        }
        for key in all_referenced_asset_keys.difference(all_keys):
            assets_defs.append(
                external_asset_from_spec(
                    AssetSpec(key=key, metadata={SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET: True})
                )
            )
        return assets_defs

    @classmethod
    def from_assets(
        cls,
        assets: Iterable[Union[AssetsDefinition, SourceAsset]],
    ) -> "AssetGraph":
        assets_defs = cls.normalize_assets(assets)

        # Build the set of AssetNodes. Each node holds key rather than object references to parent
        # and child nodes.
        dep_graph = generate_asset_dep_graph(assets_defs)

        assets_defs_by_check_key: Dict[AssetCheckKey, AssetsDefinition] = {}
        check_keys_by_asset_key: DefaultDict[AssetKey, Set[AssetCheckKey]] = defaultdict(set)
        for ad in assets_defs:
            for ck in ad.check_keys:
                check_keys_by_asset_key[ck.asset_key].add(ck)
                assets_defs_by_check_key[ck] = ad

        asset_nodes_by_key = {
            key: AssetNode(
                key=key,
                parent_keys=dep_graph["upstream"][key],
                child_keys=dep_graph["downstream"][key],
                assets_def=ad,
                check_keys=check_keys_by_asset_key[key],
            )
            for ad in assets_defs
            for key in ad.keys
        }

        return AssetGraph(
            asset_nodes_by_key=asset_nodes_by_key,
            assets_defs_by_check_key=assets_defs_by_check_key,
        )

    def get_execution_set_asset_and_check_keys(
        self, asset_or_check_key: AssetKeyOrCheckKey
    ) -> AbstractSet[AssetKeyOrCheckKey]:
        if isinstance(asset_or_check_key, AssetKey):
            return self.get(asset_or_check_key).execution_set_asset_and_check_keys
        else:  # AssetCheckKey
            assets_def = self._assets_defs_by_check_key[asset_or_check_key]
            return (
                {asset_or_check_key} if assets_def.can_subset else assets_def.asset_and_check_keys
            )

    @cached_property
    def assets_defs(self) -> Sequence[AssetsDefinition]:
        return list(
            {
                *(asset.assets_def for asset in self.asset_nodes),
                *(ad for ad in self._assets_defs_by_check_key.values()),
            }
        )

    def assets_defs_for_keys(
        self, keys: Iterable[AssetKeyOrCheckKey]
    ) -> Sequence[AssetsDefinition]:
        return list(
            {
                *[self.get(key).assets_def for key in keys if isinstance(key, AssetKey)],
                *[
                    self._assets_defs_by_check_key[key]
                    for key in keys
                    if isinstance(key, AssetCheckKey)
                ],
            }
        )

    @cached_property
    def asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return {key for ad in self.assets_defs for key in ad.check_keys}



import datetime
from typing import Optional, Sequence, Union

from dagster._annotations import experimental
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity

from ..asset_checks import AssetChecksDefinition
from ..assets import AssetsDefinition, SourceAsset
from ..events import CoercibleToAssetKey
from .shared_builder import build_freshness_checks_for_assets
from .utils import (
    DEFAULT_FRESHNESS_SEVERITY,
    DEFAULT_FRESHNESS_TIMEZONE,
)


@experimental
def build_last_update_freshness_checks(
    *,
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    lower_bound_delta: datetime.timedelta,
    deadline_cron: Optional[str] = None,
    timezone: str = DEFAULT_FRESHNESS_TIMEZONE,
    severity: AssetCheckSeverity = DEFAULT_FRESHNESS_SEVERITY,
) -> AssetChecksDefinition:
    r"""Constructs an `AssetChecksDefinition` that checks the freshness of the provided assets.

    This check passes if the asset is found to be "fresh", and fails if the asset is found to be
    "overdue". An asset is considered fresh if a record (i.e. a materialization or observation)
    exists with a timestamp greater than the "lower bound" derived from the parameters of this
    function.

    `deadline_cron` is a cron schedule that defines the deadline for when we should expect the asset
    to arrive by; if not provided, we consider the deadline to be the execution time of the check.
    `lower_bound_delta` is a timedelta that defines the lower bound for when a record could have
    arrived by. If the most recent recent record's timestamp is earlier than
    `deadline-lower_bound_delta`, the asset is considered overdue.

    Let's use two examples, one with a deadline_cron set and one without.
    Let's say I have an asset which runs on a schedule every day at 8:00 AM UTC, and usually takes
    around 45 minutes to complete. To account for operational delays, I would expect the asset to be
    done materializing every day by 9:00 AM UTC. I would set the `deadline_cron` to "0 9 \* \* \*", and
    the `lower_bound_delta` to "45 minutes". This would mean that starting at 9:00 AM, this check
    will expect a materialization record to have been created no earlier than 8:15 AM. Note that if
    the check runs at 8:59 AM, the deadline has not yet passed, and we'll instead be checking for
    the most recently passed deadline, which is yesterday.
    Let's say I have an observable source asset on a data source which I expect should never be more
    than 3 hours out of date. In this case, there's no fixed schedule for when the data should be
    updated, so I would not provide a `deadline_cron`. Instead, I would set the `lower_bound_delta`
    parameter to "3 hours". This would mean that the check will expect the most recent observation
    record to indicate data no older than 3 hours, relative to the current time, regardless of when it runs.

    The check result will contain the following metadata:
    "dagster/freshness_params": A dictionary containing the parameters used to construct the
    check
    "dagster/last_updated_time": The time of the most recent update to the asset
    "dagster/overdue_seconds": (Only present if asset is overdue) The number of seconds that the
    asset is overdue by.
    "dagster/overdue_deadline_timestamp": The timestamp that we are expecting the asset to have
    arrived by. In the case of a provided deadline_cron, this is the timestamp of the most recent
    tick of the cron schedule. In the case of no deadline_cron, this is the current time.

    Examples:
        .. code-block:: python

            # Example 1: Assets that are expected to be updated every day within 45 minutes of
            # 9:00 AM UTC
            from dagster import build_last_update_freshness_checks, AssetKey
            from .somewhere import my_daily_scheduled_assets_def

            checks_def = build_last_update_freshness_checks(
                [my_daily_scheduled_assets_def, AssetKey("my_other_daily_asset_key")],
                lower_bound_delta=datetime.timedelta(minutes=45),
                deadline_cron="0 9 * * *",
            )

            # Example 2: Assets that are expected to be updated within 3 hours of the current time
            from dagster import build_last_update_freshness_checks, AssetKey
            from .somewhere import my_observable_source_asset

            checks_def = build_last_update_freshness_checks(
                [my_observable_source_asset, AssetKey("my_other_observable_asset_key")],
                lower_bound_delta=datetime.timedelta(hours=3),
            )


    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]): The assets to
            construct checks for. All checks are incorporated into the same `AssetChecksDefinition`,
            which can be subsetted to run checks for specific assets.
        lower_bound_delta (datetime.timedelta): The check will pass if the asset was updated within
            lower_bound_delta of the current_time (no cron) or the most recent tick of the cron
            (cron provided).
        deadline_cron (Optional[str]): Defines the deadline for when we should start checking
            that the asset arrived. If not provided, the deadline is the execution time of the check.
        timezone (Optional[str]): The timezone to use when calculating freshness and deadline. If
            not provided, defaults to "UTC".

    Returns:
        AssetChecksDefinition: An `AssetChecksDefinition` object, which can execute a freshness check
            for all provided assets.
    """
    return build_freshness_checks_for_assets(
        assets=assets,
        deadline_cron=deadline_cron,
        timezone=timezone,
        severity=severity,
        lower_bound_delta=lower_bound_delta,
    )



from typing import Dict, Mapping, Sequence, Union, cast

from pydantic import BaseModel

from dagster._annotations import experimental

from .asset_check_result import AssetCheckResult
from .asset_check_spec import AssetCheckSeverity, AssetCheckSpec
from .asset_checks import AssetChecksDefinition
from .asset_key import AssetKey, CoercibleToAssetKey
from .assets import AssetsDefinition, SourceAsset
from .decorators.asset_check_decorator import multi_asset_check
from .events import AssetMaterialization
from .metadata import TableColumn, TableMetadataEntries, TableSchema


@experimental
def build_column_schema_change_checks(
    *,
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    severity: AssetCheckSeverity = AssetCheckSeverity.WARN,
) -> Sequence[AssetChecksDefinition]:
    """Returns asset checks that pass if the the column schema of the asset's latest materialization
    is the same as the column schema of the asset's previous materialization.

    Args:
        assets (Sequence[Union[AssetKey, str, AssetsDefinition, SourceAsset]]): The assets to create
            asset checks for.
        severity (AssetCheckSeverity): The severity if the check fails. Defaults to WARN.

    Returns:
        Sequence[AssetsChecksDefinition]
    """
    asset_keys = set()
    for el in assets:
        if isinstance(el, AssetsDefinition):
            asset_keys |= el.keys
        elif isinstance(el, SourceAsset):
            asset_keys.add(el.key)
        else:
            asset_keys.add(AssetKey.from_coercible(el))

    @multi_asset_check(
        specs=[
            AssetCheckSpec(
                "column_schema_change",
                asset=asset_key,
                description="Checks whether there are changes to column schema between the asset's "
                " two most recent materializations",
            )
            for asset_key in asset_keys
        ],
        can_subset=True,
    )
    def _checks(context):
        instance = context.instance
        for asset_check_key in context.selected_asset_check_keys:
            materialization_records = instance.fetch_materializations(
                limit=2, records_filter=asset_check_key.asset_key
            ).records
            if len(materialization_records) < 2:
                yield AssetCheckResult(
                    passed=True, description="The asset has been materialized fewer than 2 times"
                )
            else:
                record, prev_record = materialization_records
                metadata = cast(AssetMaterialization, record.asset_materialization).metadata
                prev_metadata = cast(
                    AssetMaterialization, prev_record.asset_materialization
                ).metadata

                column_schema = TableMetadataEntries.extract(metadata).column_schema
                prev_column_schema = TableMetadataEntries.extract(prev_metadata).column_schema
                if column_schema is None:
                    yield AssetCheckResult(
                        passed=False,
                        description="Latest materialization has no column schema metadata",
                    )
                if prev_column_schema is None:
                    yield AssetCheckResult(
                        passed=False,
                        description="Previous materialization has no column schema metadata",
                    )

                diff = TableSchemaDiff.from_table_schemas(prev_column_schema, column_schema)
                if diff.added_columns or diff.removed_columns or diff.column_type_changes:
                    description = (
                        "Column schema changed between previous and latest materialization."
                    )
                    if diff.added_columns:
                        description += f"\n\nAdded columns: {', '.join(col.name for col in diff.added_columns)}"

                    if diff.removed_columns:
                        description += f"\n\nRemoved columns: {', '.join(col.name for col in diff.removed_columns)}"

                    if diff.column_type_changes:
                        description += "\n\nColumn type changes:"
                        for col_name, type_change in diff.column_type_changes.items():
                            description += (
                                f"\n- {col_name}: {type_change.old_type} -> {type_change.new_type}"
                            )

                    yield AssetCheckResult(passed=False, severity=severity, description=description)
                else:
                    yield AssetCheckResult(
                        passed=True,
                        description="No changes to column schema between previous and latest materialization",
                    )

    return [_checks]


class TypeChange(BaseModel):
    old_type: str
    new_type: str


class TableSchemaDiff(BaseModel):
    added_columns: Sequence[TableColumn]
    removed_columns: Sequence[TableColumn]
    column_type_changes: Mapping[str, TypeChange]

    @staticmethod
    def from_table_schemas(
        old_table_schema: TableSchema, new_table_schema: TableSchema
    ) -> "TableSchemaDiff":
        old_columns_by_key = {column.name: column for column in old_table_schema.columns}
        new_columns_by_key = {column.name: column for column in new_table_schema.columns}

        added_columns = [
            column for column in new_table_schema.columns if column.name not in old_columns_by_key
        ]
        removed_columns = [
            column for column in old_table_schema.columns if column.name not in new_columns_by_key
        ]
        column_type_changes: Dict[str, TypeChange] = {}
        for name, new_column in new_columns_by_key.items():
            old_column = old_columns_by_key.get(name)
            if old_column is not None and old_column.type != new_column.type:
                column_type_changes[name] = TypeChange(
                    old_type=old_column.type, new_type=new_column.type
                )

        return TableSchemaDiff(
            added_columns=added_columns,
            removed_columns=removed_columns,
            column_type_changes=column_type_changes,
        )



import datetime
from typing import Any, Callable, Iterable, Optional, Sequence, Union, cast

import pendulum

from dagster import _check as check
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity, AssetCheckSpec
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.decorators.asset_check_decorator import (
    multi_asset_check,
)
from dagster._core.definitions.metadata import FloatMetadataValue, TimestampMetadataValue
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._core.execution.context.compute import (
    AssetCheckExecutionContext,
)
from dagster._utils.schedules import get_latest_completed_cron_tick, is_valid_cron_string

from ..assets import AssetsDefinition, SourceAsset
from ..events import AssetKey, CoercibleToAssetKey
from .utils import (
    DEADLINE_CRON_METADATA_KEY,
    FRESHNESS_PARAMS_METADATA_KEY,
    FRESHNESS_TIMEZONE_METADATA_KEY,
    LAST_UPDATED_TIMESTAMP_METADATA_KEY,
    LOWER_BOUND_DELTA_METADATA_KEY,
    OVERDUE_DEADLINE_TIMESTAMP_METADATA_KEY,
    OVERDUE_SECONDS_METADATA_KEY,
    asset_to_keys_iterable,
    ensure_no_duplicate_assets,
    get_description_for_freshness_check_result,
    get_last_updated_timestamp,
    retrieve_latest_record,
    unique_id_from_asset_keys,
)


def build_freshness_multi_check(
    asset_keys: Sequence[AssetKey],
    deadline_cron: Optional[str],
    timezone: str,
    severity: AssetCheckSeverity,
    lower_bound_delta: Optional[datetime.timedelta],
    asset_property_enforcement_lambda: Optional[Callable[[AssetsDefinition], bool]],
) -> AssetChecksDefinition:
    params_metadata: dict[str, Any] = {FRESHNESS_TIMEZONE_METADATA_KEY: timezone}
    if deadline_cron:
        params_metadata[DEADLINE_CRON_METADATA_KEY] = deadline_cron
    if lower_bound_delta:
        params_metadata[LOWER_BOUND_DELTA_METADATA_KEY] = lower_bound_delta.total_seconds()

    @multi_asset_check(
        specs=[
            AssetCheckSpec(
                "freshness_check",
                asset=asset_key,
                metadata={FRESHNESS_PARAMS_METADATA_KEY: params_metadata},
                description="Evaluates freshness for targeted asset.",
            )
            for asset_key in asset_keys
        ],
        can_subset=True,
        name=f"freshness_check_{unique_id_from_asset_keys(asset_keys)}",
    )
    def the_check(context: AssetCheckExecutionContext) -> Iterable[AssetCheckResult]:
        for check_key in context.selected_asset_check_keys:
            asset_key = check_key.asset_key
            if asset_property_enforcement_lambda:
                asset_property_enforcement_lambda(
                    context.job_def.asset_layer.asset_graph.get(check_key.asset_key).assets_def
                )
            current_timestamp = pendulum.now("UTC").timestamp()

            # Explicit call to partitions def here will be replaced with AssetSlice reference once it's available.
            partitions_def = cast(
                Optional[TimeWindowPartitionsDefinition],
                context.job_def.asset_layer.asset_graph.get(asset_key).partitions_def,
            )

            check.invariant(
                partitions_def is None
                or isinstance(partitions_def, TimeWindowPartitionsDefinition),
                "Expected partitions_def to be time-windowed.",
            )
            current_time_in_freshness_tz = pendulum.from_timestamp(current_timestamp, tz=timezone)
            latest_completed_cron_tick = get_latest_completed_cron_tick(
                deadline_cron, current_time_in_freshness_tz, timezone
            )
            deadline = check.inst_param(
                latest_completed_cron_tick or current_time_in_freshness_tz,
                "deadline",
                datetime.datetime,
            )
            if not partitions_def:
                last_completed_time_window = None
                expected_partition_key = None
            else:
                deadline_in_partitions_def_tz = pendulum.from_timestamp(
                    deadline.timestamp(), tz=partitions_def.timezone
                )
                last_completed_time_window = check.not_none(
                    partitions_def.get_prev_partition_window(deadline_in_partitions_def_tz)
                )
                expected_partition_key = partitions_def.get_partition_key_range_for_time_window(
                    last_completed_time_window
                ).start

            if lower_bound_delta:
                last_update_time_lower_bound = cast(datetime.datetime, deadline - lower_bound_delta)
            else:
                last_update_time_lower_bound = check.not_none(
                    last_completed_time_window,
                    "Expected a valid partitioned asset with a completed time window in order to determine valid freshness window.",
                ).end

            latest_record = retrieve_latest_record(
                instance=context.instance, asset_key=asset_key, partition_key=expected_partition_key
            )
            update_timestamp = get_last_updated_timestamp(latest_record, context)
            passed = (
                update_timestamp is not None
                and update_timestamp >= last_update_time_lower_bound.timestamp()
            )

            metadata = {
                FRESHNESS_PARAMS_METADATA_KEY: params_metadata,
                OVERDUE_DEADLINE_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(
                    deadline.timestamp()
                ),
            }
            if not passed:
                metadata[OVERDUE_SECONDS_METADATA_KEY] = FloatMetadataValue(
                    current_timestamp - deadline.timestamp()
                )
            if update_timestamp:
                metadata[LAST_UPDATED_TIMESTAMP_METADATA_KEY] = TimestampMetadataValue(
                    update_timestamp
                )

            yield AssetCheckResult(
                passed=passed,
                description=get_description_for_freshness_check_result(
                    passed,
                    update_timestamp,
                    last_update_time_lower_bound,
                    current_timestamp,
                    expected_partition_key,
                    record_arrival_timestamp=latest_record.timestamp if latest_record else None,
                    event_type=latest_record.event_type if latest_record else None,
                ),
                severity=severity,
                asset_key=asset_key,
                metadata=metadata,
            )

    return the_check


def build_freshness_checks_for_assets(
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    deadline_cron: Optional[str],
    timezone: str,
    severity: AssetCheckSeverity,
    asset_property_enforcement_lambda: Optional[Callable[[AssetsDefinition], bool]] = None,
    lower_bound_delta: Optional[datetime.timedelta] = None,
) -> AssetChecksDefinition:
    ensure_no_duplicate_assets(assets)
    deadline_cron = check.opt_str_param(deadline_cron, "deadline_cron")
    check.invariant(
        is_valid_cron_string(deadline_cron) if deadline_cron else True,
        "deadline_cron must be a valid cron string.",
    )
    severity = check.inst_param(severity, "severity", AssetCheckSeverity)
    timezone = check.str_param(timezone, "timezone")
    check.opt_inst_param(lower_bound_delta, "lower_bound_delta", datetime.timedelta)

    return build_freshness_multi_check(
        asset_keys=[asset_key for asset in assets for asset_key in asset_to_keys_iterable(asset)],
        deadline_cron=deadline_cron,
        timezone=timezone,
        severity=severity,
        lower_bound_delta=lower_bound_delta,
        asset_property_enforcement_lambda=asset_property_enforcement_lambda,
    )



from typing import Sequence, Union

from dagster import _check as check
from dagster._annotations import experimental
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition

from ..asset_checks import AssetChecksDefinition
from ..assets import AssetsDefinition, SourceAsset
from ..events import CoercibleToAssetKey
from .shared_builder import build_freshness_checks_for_assets
from .utils import (
    DEFAULT_FRESHNESS_SEVERITY,
    DEFAULT_FRESHNESS_TIMEZONE,
)


@experimental
def build_time_partition_freshness_checks(
    *,
    assets: Sequence[Union[SourceAsset, CoercibleToAssetKey, AssetsDefinition]],
    deadline_cron: str,
    timezone: str = DEFAULT_FRESHNESS_TIMEZONE,
    severity: AssetCheckSeverity = DEFAULT_FRESHNESS_SEVERITY,
) -> AssetChecksDefinition:
    r"""Construct an `AssetChecksDefinition` that checks the freshness of the provided assets.

    This check passes if the asset is considered "fresh" by the time that execution begins. We
    consider an asset to be "fresh" if there exists a record for the most recent partition, once
    the deadline has passed.

    `deadline_cron` is a cron schedule that defines the deadline for when we should expect the most
    recent partition to arrive by. Once a tick of the cron schedule has passed, this check will fail
    if the most recent partition has not been observed/materialized.

    Let's say I have a daily-partitioned asset which runs every day at 8:00 AM UTC, and takes around
    45 minutes to complete. To account for operational delays, I would expect the asset to be done
    materializing every day by 9:00 AM UTC. I would set the `deadline_cron` to "0 9 \* \* \*". This
    means that starting at 9:00 AM, this check will expect a record to exist for the previous day's
    partition. Note that if the check runs at 8:59 AM, the deadline has not yet passed, and we'll
    instead be checking for the most recently passed deadline, which is yesterday (meaning the
    partition representing the day before yesterday).

    The timestamp of an observation record is the timestamp indicated by the
    "dagster/last_updated_timestamp" metadata key. The timestamp of a materialization record is the
    timestamp at which that record was created.

    The check will fail at runtime if a non-time-window partitioned asset is passed in.

    The check result will contain the following metadata:
    "dagster/freshness_params": A dictionary containing the parameters used to construct the
    check.
    "dagster/last_updated_time": (Only present if the asset has been observed/materialized before)
    The time of the most recent update to the asset.
    "dagster/overdue_seconds": (Only present if asset is overdue) The number of seconds that the
    asset is overdue by.
    "dagster/overdue_deadline_timestamp": The timestamp that we are expecting the asset to have
    arrived by. This is the timestamp of the most recent tick of the cron schedule.

    Examples:
        .. code-block:: python

            from dagster import build_time_partition_freshness_checks, AssetKey
            # A daily partitioned asset that is expected to be updated every day within 45 minutes
            # of 9:00 AM UTC
            from .somewhere import my_daily_scheduled_assets_def

            checks_def = build_time_partition_freshness_checks(
                [my_daily_scheduled_assets_def],
                deadline_cron="0 9 * * *",
            )


    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]): The assets to
            construct checks for. For each passed in asset, there will be a corresponding
            constructed `AssetChecksDefinition`.
        deadline_cron (str): The check will pass if the partition time window most recently
            completed by the time of the last cron tick has been observed/materialized.
        timezone (Optional[str]): The timezone to use when calculating freshness and deadline. If
            not provided, defaults to "UTC".

    Returns:
        AssetChecksDefinition: An `AssetChecksDefinition` object, which can execute a freshness
            check for each provided asset.
    """
    return build_freshness_checks_for_assets(
        assets=assets,
        deadline_cron=deadline_cron,
        timezone=timezone,
        severity=severity,
        asset_property_enforcement_lambda=lambda assets_def: check.invariant(
            isinstance(assets_def.partitions_def, TimeWindowPartitionsDefinition),
            "Asset is not time-window partitioned.",
        ),
    )



import os
import sys
import traceback
import uuid
from types import TracebackType
from typing import Any, NamedTuple, Optional, Sequence, Tuple, Type, Union

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._serdes import whitelist_for_serdes


# mypy does not support recursive types, so "cause" has to be typed `Any`
@whitelist_for_serdes
class SerializableErrorInfo(
    NamedTuple(
        "SerializableErrorInfo",
        [
            ("message", str),
            ("stack", Sequence[str]),
            ("cls_name", Optional[str]),
            ("cause", Any),
            ("context", Any),
        ],
    )
):
    # serdes log
    # * added cause - default to None in constructor to allow loading old entries
    # * added context - default to None for similar reasons
    #
    def __new__(
        cls,
        message: str,
        stack: Sequence[str],
        cls_name: Optional[str],
        cause: Optional["SerializableErrorInfo"] = None,
        context: Optional["SerializableErrorInfo"] = None,
    ):
        return super().__new__(cls, message, stack, cls_name, cause, context)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        stack_str = "\nStack Trace:\n" + "".join(self.stack) if self.stack else ""
        cause_str = (
            "\nThe above exception was caused by the following exception:\n"
            + self.cause.to_string()
            if self.cause
            else ""
        )
        context_str = (
            "\nThe above exception occurred during handling of the following exception:\n"
            + self.context.to_string()
            if self.context
            else ""
        )

        return f"{self.message}{stack_str}{cause_str}{context_str}"

    def to_exception_message_only(self) -> "SerializableErrorInfo":
        """Return a new SerializableErrorInfo with only the message and cause set.

        This is done in cases when the context about the error should not be exposed to the user.
        """
        return SerializableErrorInfo(message=self.message, stack=[], cls_name=self.cls_name)


def _serializable_error_info_from_tb(tb: traceback.TracebackException) -> SerializableErrorInfo:
    return SerializableErrorInfo(
        # usually one entry, multiple lines for SyntaxError
        "".join(list(tb.format_exception_only())),
        tb.stack.format(),
        tb.exc_type.__name__ if tb.exc_type is not None else None,
        _serializable_error_info_from_tb(tb.__cause__) if tb.__cause__ else None,
        _serializable_error_info_from_tb(tb.__context__) if tb.__context__ else None,
    )


ExceptionInfo: TypeAlias = Union[
    Tuple[Type[BaseException], BaseException, TracebackType],
    Tuple[None, None, None],
]


def _should_redact_user_code_error() -> bool:
    return str(os.getenv("DAGSTER_REDACT_USER_CODE_ERRORS")).lower() in ("1", "true", "t")


def serializable_error_info_from_exc_info(
    exc_info: ExceptionInfo,
    # Whether to forward serialized errors thrown from subprocesses
    hoist_user_code_error: Optional[bool] = True,
) -> SerializableErrorInfo:
    """This function is used to turn an exception into a serializable object that can be passed
    across process boundaries or sent over GraphQL.

    Args:
        exc_info (ExceptionInfo): The exception info to serialize
        hoist_user_code_error (Optional[bool]): Whether to extract the inner user code error if the raised exception
            is a DagsterUserCodeProcessError. Defaults to True.
    """
    # `sys.exc_info() return Tuple[None, None, None] when there is no exception being processed. We accept this in
    # the type signature here since this function is meant to directly receive the return value of
    # `sys.exc_info`, but the function should never be called when there is no exception to process.
    exc_type, e, tb = exc_info
    additional_message = "sys.exc_info() called but no exception available to process."
    exc_type = check.not_none(exc_type, additional_message=additional_message)
    e = check.not_none(e, additional_message=additional_message)
    tb = check.not_none(tb, additional_message=additional_message)

    from dagster._core.errors import DagsterUserCodeExecutionError, DagsterUserCodeProcessError

    if isinstance(e, DagsterUserCodeExecutionError) and _should_redact_user_code_error():
        error_id = str(uuid.uuid4())

        sys.stderr.write(f"Error occurred during user code execution, error ID {error_id}.\n")
        sys.stderr.write(
            _serializable_error_info_from_tb(
                traceback.TracebackException(exc_type, e, tb)
            ).to_string(),
        )
        return SerializableErrorInfo(
            message=(
                f"Error occurred during user code execution, error ID {error_id}. "
                "The error has been masked to prevent leaking sensitive information. "
                "Search in logs for this error ID for more details."
            ),
            stack=[],
            cls_name="DagsterRedactedUserCodeError",
            cause=None,
            context=None,
        )

    if (
        hoist_user_code_error
        and isinstance(e, DagsterUserCodeProcessError)
        and len(e.user_code_process_error_infos) == 1
    ):
        return e.user_code_process_error_infos[0]
    else:
        tb_exc = traceback.TracebackException(exc_type, e, tb)
        return _serializable_error_info_from_tb(tb_exc)



from typing import Any, Dict, Optional

import pydantic
from pydantic import BaseModel
from typing_extensions import Self


class DagsterModel(BaseModel):
    """Standardizes on Pydantic settings that are stricter than the default.
    - Frozen, to avoid complexity caused by mutation.
    - extra=forbid, to avoid bugs caused by accidentally constructing with the wrong arguments.
    - arbitrary_types_allowed, to allow non-model class params to be validated with isinstance.
    """

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    class Config:
        extra = "forbid"
        frozen = True
        arbitrary_types_allowed = True

    def model_copy(self, *, update: Optional[Dict[str, Any]] = None) -> Self:
        if pydantic.__version__ >= "2":
            return super().model_copy(update=update)  # type: ignore
        else:
            return super().copy(update=update)
